#!/usr/bin/env sage
# vim: ft=python
#
# Forfatter: John Aslak Wee Kleven
# Lisens: GPL 3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html)

import re

# Gram-Schmidt, med ortonormalisering.
def gram_schmidt(S):
	out = []

	out.append(S[0] / sqrt(ip(S[0], S[0])))

	for k in range(1,len(S)):
		tmp = S[k]

		for u in out:
			nevner = ip(u, u)

			if nevner == 0:
				continue

			if numerical:
				# If simplifying down to numerical
				tmp -= N(ip( S[k], u )) * u
			else:
				tmp -= ip( S[k], u ) * u

		kvdrlen = ip(tmp, tmp)

		if kvdrlen == 0:
			continue

		out.append(tmp / sqrt(kvdrlen))

	return out

# Teorem: Gitt et indreproduktrom V, v i V og en ortonormal mengde S, er det
# nærmeste elementet i span(S) til v, lik sum_{x i S} <v,x> x.

def approximate(basis, func, str_output=0):
	out = 0
	for k in basis:
		out += ip(func,k) * k

	# Make the symbolic expression slightly prettier
	if not numerical:
		out = out.simplify()

	if not str_output:
		return str(out)

	else:

		# Do string handling to make geogebra happy
		# note: e06 has exponent -6.
		out = re.sub('e-0([1-9]+)', '*10^(-\\1)', str(out))
		out = re.sub('e-([1-9]+)', '*10^(\\1)', str(out))
		return out


################################################################################

numerical = 1 # calculate exactly, or numerically?
approximated = exp(x)
interval = [0, 2]
basis = [1, x, x^2]

# Inner product; global variable
if numerical: ip = lambda f,g: numerical_integral(f*g,interval[0],interval[1])[0]
else: ip = lambda f,g: integral(f*g,x,interval[0],interval[1])

orig_basis = deepcopy(basis)
basis = gram_schmidt(basis)

prox = approximate(basis, approximated, str_output=1)
print(f"Approximation of {approximated} of {orig_basis}")
print(f"on the interval [{interval[0]}, {interval[1]}]")
print(prox)
print(f"The orthonormal basis: {basis}")


# Lag et plott

f(x) = prox
g(x) = approximated

print(f"Distance in the inner product metric: {N(sqrt(ip(f-g,f-g)))}")

k = plot( f(x), x, interval[0], interval[1], color="blue") + \
	plot( g(x), x, interval[0], interval[1], color="red")

i = plot( f(x), x, interval[0], interval[1], color="blue" ) + plot(g(x),x,interval[0],interval[1],color="red")

NAME="func_approx.png"
print(f"Plot saved to {NAME}.")
i.save(NAME, dpi=500)
